"""Provides the basic protocol class (the abstract interface) for a channel."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing import Any, Mapping, Sequence

    from choreographer.protocol import BrowserResponse


class ChannelInterface(Protocol):
    """Defines the basic interface of a channel."""

    # Not sure I like the obj type
    def write_json(self, obj: Mapping[str, Any]) -> tuple[float, float]:
        ...
        # """
        # Accept an object and send it doesnt the channel serialized.
        #
        # Args:
        #   obj: the object to send to the browser.
        #
        # """

    def read_jsons(
        self,
        *,
        blocking: bool = True,
    ) -> tuple[Sequence[BrowserResponse], float]:
        ...
        # """
        # Read all available jsons in the channel and returns a list of complete ones.
        #
        # Args:
        #   blocking: should this method block on read or return immediately.
        # """

    def close(self) -> None:
        ...
        # """Close the channel."""

    def open(self) -> None:
        ...
        # """Open the channel."""

    def is_ready(self) -> bool:
        ...
        # """Return true if comm channel is active."""


# Can't docstring protocols! EW!



from __future__ import annotations

import os
import platform
from dataclasses import dataclass

chrome_names = (
    "chrome",
    "Chrome",
    "google-chrome",
    "google-chrome-stable",
    "Chrome.app",
    "Google Chrome",
    "Google Chrome.app",
    "Google Chrome for Testing",
)

chromium_names = ("chromium", "chromium-browser", "Chromium")

edge_names = (
    "msedge",
    "Microsoft Edge",
    "microsoft-edge",
    "microsoft-edge-beta",
    "microsoft-edge-dev",
)

brave_names = ("brave", "Brave Browser", "brave-browser")

vivaldi_names = ("vivaldi", "Vivaldi")


if platform.system() == "Windows":
    _windows_app_dirs = [
        _p
        for _p in [
            os.environ.get("PROGRAMFILES", ""),
            os.environ.get("PROGRAMFILES(X86)", ""),
            os.environ.get("LOCALAPPDATA", ""),
        ]
        if _p
    ]
    _chrome_suffix = r"\Google\Chrome\Application\chrome.exe"
    typical_chrome_paths = tuple(_p + _chrome_suffix for _p in _windows_app_dirs)

    _chromium_suffix = r"\Chromium\Application\chrome.exe"
    typical_chromium_paths = tuple(_p + _chromium_suffix for _p in _windows_app_dirs)

    _edge_suffix = r"\Microsoft\Edge\Application\msedge.exe"
    typical_edge_paths = tuple(_p + _edge_suffix for _p in _windows_app_dirs)

    _brave_suffix = r"\BraveSoftware\Brave-Browser\Application\brave.exe"
    typical_brave_paths = tuple(_p + _edge_suffix for _p in _windows_app_dirs)

    _vivaldi_suffix = r"\Vivaldi\Application\vivaldi.exe"
    typical_vivaldi_paths = tuple(_p + _brave_suffix for _p in _windows_app_dirs)

elif platform.system() == "Linux":
    typical_chrome_paths = (
        "/usr/bin/google-chrome-stable",
        "/usr/bin/google-chrome",
        "/usr/bin/chrome",
    )
    typical_chromium_paths = (
        "/usr/bin/chromium",
        "/usr/bin/chromium-browser",
    )
    typical_edge_paths = (
        "/usr/bin/microsoft-edge",
        "/usr/bin/microsoft-edge-beta",
        "/usr/bin/microsoft-edge-dev",
    )
    typical_brave_paths = (
        "/usr/bin/brave-browser",
        "/opt/brave.com/brave/brave-browser",
    )
    typical_vivaldi_paths = ("/usr/bin/vivaldi",)

else:  # assume mac, or system == "Darwin"
    typical_chrome_paths = (
        "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
    )
    typical_chromium_paths = ("/Applications/Chromium.app/Contents/MacOS/Chromium",)

    typical_edge_paths = (
        "/Applications/Microsoft Edge.app/Contents/MacOS/Microsoft Edge",
    )

    typical_brave_paths = (
        "/Applications/Brave Browser.app/Contents/MacOS/Brave Browser",
    )
    typical_vivaldi_paths = ("/Applications/Vivaldi.app/Contents/MacOS/Vivaldi",)


@dataclass(frozen=True)
class BrowserInfo:
    __slots__ = ("exe_names", "ms_prog_id", "typical_paths")
    exe_names: tuple[str, ...]
    ms_prog_id: str
    typical_paths: tuple[str, ...]


chromium_based_browsers = {
    "chrome": BrowserInfo(chrome_names, "ChromeHTML", typical_chrome_paths),
    "chromium": BrowserInfo(chromium_names, "ChromiumHTML", typical_chromium_paths),
    "edge": BrowserInfo(edge_names, "MSEdgeHTM", typical_edge_paths),
    "brave": BrowserInfo(brave_names, "BraveHTML", typical_brave_paths),
    "vivaldi": BrowserInfo(vivaldi_names, "VivaldiHTM", typical_vivaldi_paths),
}



import os
import platform
import stat
import sys
from pathlib import Path

import pytest

from choreographer.browsers import chromium
from choreographer.cli._cli_utils import get_chrome_download_path


def test_internal(tmp_path):
    if platform.system() == "Windows":
        pytest.skip("Windows hard to trick about PATH.")
    _o = str(os.environ["PATH"])
    os.environ["PATH"] = str(tmp_path)
    print(os.environ["PATH"])
    names = ["chrome", "chromium", "msedge", "brave", "vivaldi"]
    paths = []

    for n in names:
        p = tmp_path / (n)
        p.touch()
        p.chmod(p.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
        paths.append(p)

    for _p, _n in zip(paths, names):
        _r = chromium.Chromium.find_browser(skip_local=True, skip_typical=True)
        assert _r
        assert Path(_r).stem == _n
        _p.unlink()

    os.environ["PATH"] = _o


def test_canary():
    # This ensures we are finding the local install
    if os.getenv("TEST_SYSTEM_BROWSER"):
        pytest.skip("Okay, no need to test for local.")
    _r = chromium.Chromium.find_browser(skip_local=False, skip_typical=True)
    print(sys.path, file=sys.stderr)
    print(f"found {_r}", file=sys.stderr)
    print(f"download path: {get_chrome_download_path()}", file=sys.stderr)
    assert _r
    assert Path(_r) == get_chrome_download_path()
    # if _r != get_chrome_download_path():



import asyncio

import logistro
import pytest

from choreographer.protocol.devtools_async_helpers import (
    create_and_wait,
    execute_js_and_wait,
    navigate_and_wait,
)

pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


# Errata: don't use data urls, whether or not they load is variable
# depends on how long chrome has been open for, how they were entered,
# etc


@pytest.mark.asyncio
async def test_create_and_wait(browser):
    """Test create_and_wait with both valid data URL and blank URL."""
    _logger.info("testing create_and_wait...")

    # Count tabs before
    initial_tab_count = len(browser.tabs)

    # Create a simple HTML page as a data URL
    data_url = "https://www.example.com"

    # Test 1: Create tab with data URL - should succeed
    tab1 = await create_and_wait(browser, url=data_url, timeout=5.0)
    assert tab1 is not None

    # Verify the page loaded correctly using execute_js_and_wait
    result = await execute_js_and_wait(tab1, "window.location.href", timeout=5.0)
    location = result["result"]["result"]["value"]
    assert location.startswith(data_url)

    # Test 2: Create tab without URL - should succeed (blank page)
    tab2 = await create_and_wait(browser, url="", timeout=5.0)
    assert tab2 is not None

    # Verify we have 2 more tabs
    final_tab_count = len(browser.tabs)
    assert final_tab_count == initial_tab_count + 2

    # Test 3: Create tab with bad URL that won't load - should timeout
    with pytest.raises(asyncio.TimeoutError):
        await create_and_wait(browser, url="http://192.0.2.1:9999", timeout=0.5)


@pytest.mark.asyncio
async def test_navigate_and_wait(browser):
    """Test navigate_and_wait with both valid data URL and bad URL."""
    _logger.info("testing navigate_and_wait...")
    # Create two blank tabs first
    tab = await browser.create_tab("")

    # navigating to dataurls seems to be fine right now,
    # but if one day you have an error here,
    # change to the strategy above

    # Create a data URL with identifiable content
    html_content1 = "<html><body><h1>Navigation Test 1</h1></body></html>"
    data_url1 = f"data:text/html,{html_content1}"

    html_content2 = "<html><body><h1>Navigation Test 2</h1></body></html>"
    data_url2 = f"data:text/html,{html_content2}"

    # Test 1: Navigate first tab to valid data URL - should succeed
    result_tab1 = await navigate_and_wait(tab, url=data_url1, timeout=5.0)
    assert result_tab1 is tab

    # Verify the navigation succeeded using execute_js_and_wait
    result = await execute_js_and_wait(tab, "window.location.href", timeout=5.0)
    location = result["result"]["result"]["value"]
    assert location.startswith("data:text/html")

    # Test 2: Navigate second tab to another valid data URL - should succeed
    result_tab2 = await navigate_and_wait(tab, url=data_url2, timeout=5.0)
    assert result_tab2 is tab

    # Verify the navigation succeeded
    result = await execute_js_and_wait(tab, "window.location.href", timeout=5.0)
    location = result["result"]["result"]["value"]
    assert location.startswith("data:text/html")
    # Test 3: Navigate to bad URL that won't load - should timeout
    with pytest.raises(asyncio.TimeoutError):
        await navigate_and_wait(tab, url="http://192.0.2.1:9999", timeout=0.5)



"""Defaults used when arguments not supplied."""

from pathlib import Path

from platformdirs import PlatformDirs

default_download_path = (
    Path(
        PlatformDirs("choreographer", "plotly").user_data_dir,
    )
    / "deps"
)
"""The path where we download chrome if no path argument is supplied."""



from __future__ import annotations

import os
import platform
import subprocess

import logistro

if (_system := platform.system()) != "Windows":
    import signal

_logger = logistro.getLogger(__name__)


def kill(process: subprocess.Popen[bytes] | subprocess.Popen[str]) -> None:
    if _system == "Windows":
        subprocess.call(  # noqa: S603, false positive, input fine
            ["taskkill", "/F", "/T", "/PID", str(process.pid)],  # noqa: S607 windows full path...
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            timeout=6,
        )
        return

    try:
        os.killpg(
            process.pid,
            signal.SIGTERM,  # type: ignore[reportPossiblyUnboundVariable]
        )
    except ProcessLookupError:
        process.terminate()
    _logger.debug("Called terminate (a light kill).")
    try:
        process.wait(timeout=6)
    except subprocess.TimeoutExpired:
        _logger.debug("Calling kill (a heavy kill).")
        try:
            os.killpg(
                process.pid,
                signal.SIGKILL,  # type: ignore[reportPossiblyUnboundVariable]
            )
        except ProcessLookupError:
            process.kill()



"""Provides the basic protocol class (the abstract base) for a protocol."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    import logging
    from pathlib import Path
    from typing import Any, Mapping, MutableMapping, Sequence

    from choreographer.channels._interface_type import ChannelInterface


class BrowserImplInterface(Protocol):
    """Defines the basic interface of a channel."""

    path: str | Path | None
    """The OS path to the operating system."""

    @classmethod
    def find_browser(
        cls,
        *,
        skip_local: bool,
        skip_typical: bool = False,
    ) -> str | None: ...

    ### Tries to find a working copy of itself, using our OS methods
    ### as well as its own methods. See the Chromium implementation.

    @classmethod
    def logger_parser(
        cls,
        record: logging.LogRecord,
        _old: MutableMapping[str, Any],
    ) -> bool: ...

    # This method will be used as the `filter()` method on the `logging.Filter()`
    # attached to all incoming logs from the browser process.
    # The log will be entirely ignored if return is False.

    def __init__(
        self,
        channel: ChannelInterface,
        path: Path | str | None = None,
        **kwargs: Any,
    ) -> None: ...
    def pre_open(self) -> None: ...
    def get_popen_args(self) -> Mapping[str, Any]: ...
    def get_cli(self) -> Sequence[str]: ...
    def get_env(self) -> MutableMapping[str, str]: ...
    def clean(self) -> None: ...
    def is_isolated(self) -> bool: ...



"""Async helper functions for common Chrome DevTools Protocol patterns."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import logistro

if TYPE_CHECKING:
    from choreographer import Browser, Tab
    from choreographer.protocol.devtools_async import Session

    from . import BrowserResponse

_logger = logistro.getLogger(__name__)

# Abit about the mechanics of chrome:
# Whether or not a Page.loadEventFired event fires is a bit
# racey. Optimistically, it's buffered and fired after subscription
# even if the event happened in the past.
# Doesn't seem to always work out that way, so we also use
# javascript to create a "loaded" event, but for the case
# where we need to timeout- loading a page that never resolves,
# the browser might actually load an about:blank instead and then
# fire the event, misleading the user, so we check the url.


async def _check_document_ready(session: Session, url: str) -> BrowserResponse:
    return await session.send_command(
        "Runtime.evaluate",
        params={
            "expression": """
                new Promise((resolve) => {
                    if (
                        (document.readyState === 'complete') &&
                        (window.location==`"""  # CONCATENATE!
            f"{url!s}"
            """`)
                    ){
                        resolve("Was complete");
                    } else {
                        window.addEventListener(
                            'load', () => resolve("Event loaded")
                        );
                    }
                })
            """,
            "awaitPromise": True,
            "returnByValue": True,
        },
    )


async def create_and_wait(
    browser: Browser,
    url: str = "",
    *,
    timeout: float = 30.0,
) -> Tab:
    """
    Create a new tab and wait for it to load.

    Args:
        browser: Browser instance
        url: URL to navigate to (default: blank page)
        timeout: Seconds to wait for page load (default: 30.0)

    Returns:
        The created Tab

    """
    _logger.debug("Creating tab")
    tab = await browser.create_tab(url)
    _logger.debug("Creating session")
    temp_session = await tab.create_session()

    try:
        _logger.debug("Subscribing to loadEven and enabling events.")
        load_future = temp_session.subscribe_once("Page.loadEventFired")
        await temp_session.send_command("Page.enable")
        await temp_session.send_command("Runtime.enable")

        if url:
            try:
                # JavaScript evaluation to check if document is loaded
                js_ready_future = asyncio.create_task(
                    _check_document_ready(temp_session, url),
                )
                _logger.debug(f"Starting wait: timeout={timeout}")
                # Race between the two methods: first one to complete wins
                done, pending = await asyncio.wait(
                    [
                        load_future,
                        js_ready_future,
                    ],
                    return_when=asyncio.FIRST_COMPLETED,
                    timeout=timeout,
                )
                _logger.debug(f"Finish wait, is done? {bool(done)}")

                for task in pending:
                    _logger.debug(f"Cancelling: {task}")
                    task.cancel()

                if not done:
                    _logger.debug("Timeout waiting for js or event")
                    raise asyncio.TimeoutError(  # noqa: TRY301
                        "Page load timeout",
                    )
                else:
                    _logger.debug(f"Task which finished: {done}")

            except (
                asyncio.TimeoutError,
                asyncio.CancelledError,
                TimeoutError,
            ) as e:
                # Stop the page load when timeout occurs
                _logger.debug("Need to stop page loading, error.", exc_info=e)
                await temp_session.send_command("Page.stopLoading")
                raise
    finally:
        _logger.debug("Closing session")
        await tab.close_session(temp_session.session_id)

    _logger.debug("Returning tab.")
    return tab


async def navigate_and_wait(
    tab: Tab,
    url: str,
    *,
    timeout: float = 30.0,
) -> Tab:
    """
    Navigate an existing tab to a URL and wait for it to load.

    Args:
        tab: Tab to navigate
        url: URL to navigate to
        timeout: Seconds to wait for page load (default: 30.0)

    Returns:
        The Tab after navigation completes

    """
    temp_session = await tab.create_session()

    try:
        # Subscribe BEFORE enabling domains to avoid race condition
        load_future = temp_session.subscribe_once("Page.loadEventFired")
        await temp_session.send_command("Page.enable")
        await temp_session.send_command("Runtime.enable")
        try:

            async def _freezers() -> None:
                # If no resolve, will freeze
                await temp_session.send_command("Page.navigate", params={"url": url})
                # Can freeze if resolve bad
                await load_future

            await asyncio.wait_for(_freezers(), timeout=timeout)
        except (asyncio.TimeoutError, asyncio.CancelledError, TimeoutError):
            # Stop the navigation when timeout occurs
            await temp_session.send_command("Page.stopLoading")
            raise
    finally:
        await tab.close_session(temp_session.session_id)

    return tab


async def execute_js_and_wait(
    tab: Tab,
    expression: str,
    *,
    timeout: float = 30.0,
) -> BrowserResponse:
    """
    Execute JavaScript in a tab and return the result.

    Args:
        tab: Tab to execute JavaScript in
        expression: JavaScript expression to evaluate
        timeout: Seconds to wait for execution (default: 30.0)

    Returns:
        Response dict from Runtime.evaluate with 'result' and optional
        'exceptionDetails'

    """
    temp_session = await tab.create_session()

    try:
        await temp_session.send_command("Page.enable")
        await temp_session.send_command("Runtime.enable")

        response = await asyncio.wait_for(
            temp_session.send_command(
                "Runtime.evaluate",
                params={
                    "expression": expression,
                    "awaitPromise": True,
                    "returnByValue": True,
                },
            ),
            timeout=timeout,
        )

        return response
    finally:
        await tab.close_session(temp_session.session_id)



from __future__ import annotations

try:
    from datetime import UTC, datetime  #  type: ignore [attr-defined]
except ImportError:
    from datetime import datetime, timezone

    UTC = timezone.utc

import json
from typing import TYPE_CHECKING

import choreographer.channels._wire as wire
import logistro
import numpy as np
import pytest
from choreographer.channels import register_custom_encoder

if TYPE_CHECKING:
    from typing import Any

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")


_timestamp = datetime(1970, 1, 1, tzinfo=UTC)

data = [1, 2.00, 3, float("nan"), float("inf"), float("-inf"), _timestamp]
expected_message = b'[1, 2.0, 3, null, null, null, "1970-01-01T00:00:00+00:00"]'
converted_type = [int, float, int, type(None), type(None), type(None), str]

_logger = logistro.getLogger(__name__)


async def test_custom_encoder():
    class NonsenseEncoder(json.JSONEncoder):
        def iterencode(
            self,
            o: Any,
            _one_shot: bool = False,  # noqa: FBT001, FBT002
        ) -> Any:
            _ = o
            yield "Test Passed."

        def encode(self, o: Any) -> Any:
            _ = o
            return "Test Passed."

    register_custom_encoder(NonsenseEncoder)
    message = wire.serialize(data)
    assert message == b"Test Passed."
    register_custom_encoder(None)
    message = wire.serialize(data)
    assert message == expected_message


@pytest.mark.asyncio
async def test_de_serialize():
    _logger.info("testing...")
    message = wire.serialize(data)
    assert message == expected_message
    obj = wire.deserialize(message.decode())
    assert len(obj) == len(converted_type)
    for o, t in zip(obj, converted_type):
        assert isinstance(o, t)
    message_np = wire.serialize(np.array(data))
    assert message_np == expected_message
    obj_np = wire.deserialize(message_np.decode())
    assert len(obj_np) == len(converted_type)
    for o, t in zip(obj_np, converted_type):
        assert isinstance(o, t)



import asyncio

import logistro
import pytest

import choreographer as choreo
from choreographer import errors
from choreographer.protocol import devtools_async

# We no longer use live URLs to as not depend on the network

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


@pytest.mark.asyncio
async def test_create_and_close_tab(browser):
    _logger.info("testing...")
    tab = await browser.create_tab("")
    assert isinstance(tab, choreo.Tab)
    assert tab.target_id in browser.tabs
    await browser.close_tab(tab)
    assert tab.target_id not in browser.tabs


@pytest.mark.asyncio
async def test_create_and_close_session(browser):
    _logger.info("testing...")
    with pytest.warns(errors.ExperimentalFeatureWarning):
        session = await browser.create_session()
    assert isinstance(session, devtools_async.Session)
    assert session.session_id in browser.sessions
    await browser.close_session(session)
    assert session.session_id not in browser.sessions


# Along with testing, this could be repurposed as well to diagnose
# This deserves some thought re. difference between write_json and send_command
@pytest.mark.asyncio
async def test_broker_write_json(browser):
    _logger.info("testing...")
    # Test valid request with correct id and method
    response = await browser._broker.write_json(  # noqa: SLF001
        {"id": 0, "method": "Target.getTargets"},
    )
    assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 I like this assertion

    # Test invalid method name should return error
    response = await browser._broker.write_json(  # noqa: SLF001
        {"id": 2, "method": "dkadklqwmd"},
    )
    assert "error" in response

    # Test missing 'id' key
    with pytest.raises(
        errors.MissingKeyError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"method": "Target.getTargets"},
        )

    # Test missing 'method' key
    with pytest.raises(
        errors.MissingKeyError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": 1},
        )

    # Test empty dictionary
    with pytest.raises(
        errors.MissingKeyError,
    ):
        await browser._broker.write_json({})  # noqa: SLF001

    # Test invalid parameter in the message
    with pytest.raises(
        RuntimeError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": 0, "method": "Target.getTargets", "invalid_parameter": "kamksamdk"},
        )

    # Test int method should return error
    with pytest.raises(
        errors.MessageTypeError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": 3, "method": 12345},
        )

    # Test non-integer id should return error
    with pytest.raises(
        errors.MessageTypeError,
    ):
        await browser._broker.write_json(  # noqa: SLF001
            {"id": "string", "method": "Target.getTargets"},
        )


@pytest.mark.asyncio
async def test_browser_send_command(browser):
    _logger.info("testing...")
    # Test valid request with correct command
    response = await browser.send_command(command="Target.getTargets")
    assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 I like this assertion

    # Test invalid method name should return error
    response = await browser.send_command(command="dkadklqwmd")
    assert "error" in response

    # Test int method should return error
    with pytest.raises(
        errors.MessageTypeError,
    ):
        await browser.send_command(command=12345)


@pytest.mark.asyncio
async def test_browser_send_command_with_perf(browser):
    _logger.info("testing...")
    perfs = []

    # Run multiple commands and collect perf data
    for _ in range(3):
        response, perf = await browser.send_command(
            command="Target.getTargets",
            with_perf=True,
        )
        assert "result" in response and "targetInfos" in response["result"]  # noqa: PT018 I like this assertion

        # Validate perf is a tuple of 3 floats
        assert isinstance(perf, tuple)
        assert all(isinstance(t, float) for t in perf)

        # Validate timing makes sense (write_start <= write_end <= read_end)
        write_start, write_end, read_end = perf
        assert write_start <= write_end
        assert write_start <= read_end
        # read_end is tough because it happens on a different processor
        # and it might actually be somewhat skewed

        perfs.append(perf)
        await asyncio.sleep(0.1)

    # Verify each perf tuple is distinct and sequential
    for i in range(len(perfs) - 1):
        _, _, prev_read_end = perfs[i]
        next_write_start, _, _ = perfs[i + 1]

        # Next command should start after previous command ended
        assert prev_read_end <= next_write_start, "Commands should be sequential"


@pytest.mark.asyncio
async def test_populate_targets(browser):
    _logger.info("testing...")
    await browser.send_command(command="Target.createTarget", params={"url": ""})
    await browser.populate_targets()
    assert len(browser.tabs) >= 1


@pytest.mark.asyncio
async def test_get_tab(browser):
    _logger.info("testing...")
    await browser.create_tab("")
    assert browser.get_tab() == next(iter(browser.tabs.values()))
    await browser.create_tab()
    await browser.create_tab("")
    assert browser.get_tab() == next(iter(browser.tabs.values()))



import io
import zipfile
from unittest.mock import patch

import pytest

from choreographer.cli._cli_utils import get_chrome_sync


@pytest.fixture
def mock_last_known_good_json():
    """Mock the last known good chrome version."""
    return {
        "version": "135.0.7011.0",
        "revision": "1418433",
        "downloads": {
            "chrome": [
                {"platform": "linux64", "url": "https://example.com/linux64.zip"},
                {"platform": "mac-arm64", "url": "https://example.com/mac-arm64.zip"},
                {"platform": "win64", "url": "https://example.com/win64.zip"},
            ],
        },
    }


# Thanks Claude!
def test_get_chrome_sync_download_behavior(
    tmp_path,
    mock_last_known_good_json,
):
    """Test chrome download/skip behavior: existing, force, version change."""
    exe_path = tmp_path / "chrome-linux64" / "chrome"
    version_tag = tmp_path / "version_tag.txt"

    # Setup: create a fake existing installation with matching version
    exe_path.parent.mkdir(parents=True)
    exe_path.write_text("old chrome")
    version_tag.write_text("135.0.7011.0\n1418433")

    # Mock the URL download
    def create_mock_zip_response():
        # Create a fresh zip buffer each time
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w") as zf:
            zf.writestr("chrome-linux64/chrome", "newly downloaded chrome")
        zip_buffer.seek(0)
        # Return BytesIO directly - it already has all file-like methods needed
        return zip_buffer

    # Patch json.load to return our mock data (avoid broad Path.read_text patch)
    with patch("json.loads", return_value=mock_last_known_good_json), patch(
        "urllib.request.urlopen",
        side_effect=lambda url: create_mock_zip_response(),  # noqa: ARG005
    ):
        # a) First call without force - should return existing, no download
        result = get_chrome_sync(arch="linux64", path=tmp_path, force=False)
        assert result == exe_path
        assert exe_path.read_text() == "old chrome"  # Should still be old

        # b) Call with force=True - should download and replace
        result = get_chrome_sync(arch="linux64", path=tmp_path, force=True)
        assert result == exe_path
        assert exe_path.read_text() == "newly downloaded chrome"
        assert version_tag.read_text() == "135.0.7011.0\n1418433"

        # c) Call again without force - should return existing, no re-download
        # (Modify the file to verify it doesn't get re-downloaded)
        exe_path.write_text("manually modified chrome")
        result = get_chrome_sync(arch="linux64", path=tmp_path, force=False)
        assert result == exe_path
        assert exe_path.read_text() == "manually modified chrome"

        # d) Change version on disk - should trigger download (version mismatch)
        version_tag.write_text("999.0.0.0\n999999")  # Old/different version
        result = get_chrome_sync(arch="linux64", path=tmp_path, force=False)
        assert result == exe_path
        assert exe_path.read_text() == "newly downloaded chrome"
        assert version_tag.read_text() == "135.0.7011.0\n1418433"  # Updated to new



from __future__ import annotations

import argparse
import asyncio
import json
import platform
import shutil
import sys
import urllib.request
import warnings
import zipfile
from functools import partial
from pathlib import Path

import logistro

from choreographer.cli.defaults import default_download_path

_logger = logistro.getLogger(__name__)

_chrome_for_testing_url = "https://googlechromelabs.github.io/chrome-for-testing/known-good-versions-with-downloads.json"

supported_platform_strings = ["linux64", "win32", "win64", "mac-x64", "mac-arm64"]


def get_google_supported_platform_string() -> str | None:
    arch_size_detected = "64" if sys.maxsize > 2**32 else "32"
    arch_detected = "arm" if platform.processor() == "arm" else "x"

    chrome_platform_detected: str | None = None
    if platform.system() == "Windows":
        chrome_platform_detected = "win" + arch_size_detected
    elif platform.system() == "Linux":
        chrome_platform_detected = "linux" + arch_size_detected
    elif platform.system() == "Darwin":
        chrome_platform_detected = "mac-" + arch_detected + arch_size_detected

    platform_string = ""
    if chrome_platform_detected in supported_platform_strings:
        platform_string = chrome_platform_detected

    return platform_string


def get_chrome_download_path() -> Path | None:
    _chrome_platform_detected = get_google_supported_platform_string()

    if not _chrome_platform_detected:
        return None

    _default_exe_path = default_download_path
    _default_exe_path.mkdir(parents=True, exist_ok=True)

    if platform.system().startswith("Linux"):
        _default_exe_path = (
            default_download_path / f"chrome-{_chrome_platform_detected}" / "chrome"
        )
    elif platform.system().startswith("Darwin"):
        _default_exe_path = (
            default_download_path
            / f"chrome-{_chrome_platform_detected}"
            / "Google Chrome for Testing.app"
            / "Contents"
            / "MacOS"
            / "Google Chrome for Testing"
        )
    elif platform.system().startswith("Win"):
        _default_exe_path = (
            default_download_path / f"chrome-{_chrome_platform_detected}" / "chrome.exe"
        )

    return _default_exe_path


# https://stackoverflow.com/questions/39296101/python-zipfile-removes-execute-permissions-from-binaries
class _ZipFilePermissions(zipfile.ZipFile):
    def _extract_member(self, member, targetpath, pwd):  # type: ignore [no-untyped-def]
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        path = super()._extract_member(member, targetpath, pwd)  # type: ignore [misc]
        # High 16 bits are os specific (bottom is st_mode flag)
        attr = member.external_attr >> 16
        if attr != 0:
            Path(path).chmod(attr)
        return path


def get_chrome_sync(  # noqa: C901, PLR0912
    arch: str | None = None,
    i: int | None = None,
    path: str | Path = default_download_path,
    *,
    verbose: bool = False,
    force: bool = False,
) -> Path | str:
    """Download chrome synchronously: see `get_chrome()`."""
    if isinstance(path, str):
        path = Path(path)

    arch = arch or get_google_supported_platform_string()
    if not arch:
        raise RuntimeError(
            "You must specify an arch, one of: "
            f"{', '.join(supported_platform_strings)}. "
            f"Detected {arch} is not supported.",
        )

    if i:
        _logger.info("Loading chrome from list")
        raw_json = urllib.request.urlopen(  # noqa: S310 audit url for schemes
            _chrome_for_testing_url,
        ).read()
        browser_list = json.loads(
            raw_json,
        )
        version_obj = browser_list["versions"][i]
        raw_json = json.dumps(version_obj)
    else:
        _logger.info("Using last known good version of chrome")
        raw_json = (
            Path(__file__).resolve().parent.parent
            / "resources"
            / "last_known_good_chrome.json"
        ).read_text()
        version_obj = json.loads(
            raw_json,
        )
    version_string = f"{version_obj['version']}\n{version_obj['revision']}"
    chromium_sources = version_obj["downloads"]["chrome"]

    for src in chromium_sources:
        if src["platform"] == arch:
            url = src["url"]
            break
    else:
        raise RuntimeError(
            "You must specify an arch, one of: "
            f"{', '.join(supported_platform_strings)}. "
            f"{arch} is not supported.",
        )

    if verbose:
        print(raw_json)  # noqa: T201 allow print in cli
    version_tag = path / "version_tag.txt"

    path.mkdir(parents=True, exist_ok=True)

    if arch.startswith("linux"):
        exe_path = path / f"chrome-{arch}" / "chrome"
    elif arch.startswith("mac"):
        exe_path = (
            path
            / f"chrome-{arch}"
            / "Google Chrome for Testing.app"
            / "Contents"
            / "MacOS"
            / "Google Chrome for Testing"
        )
    elif arch.startswith("win"):
        exe_path = path / f"chrome-{arch}" / "chrome.exe"
    else:
        raise RuntimeError("Couldn't calculate exe_name, unsupported architecture.")

    if (
        exe_path.exists()
        and version_tag.is_file()
        and version_tag.read_text() == version_string
        and not force
    ):
        return exe_path
    else:
        if exe_path.exists():  # delete it
            if exe_path.is_dir():
                shutil.rmtree(exe_path)
            else:
                exe_path.unlink()
            # It really should always be a dir but in testing we fake it
        if version_tag.exists():  # delete it
            version_tag.unlink()

    # Download
    zip_path = path / "chrome.zip"
    with urllib.request.urlopen(url) as response, zip_path.open("wb") as out_file:  # noqa: S310 audit url
        shutil.copyfileobj(response, out_file)
    with _ZipFilePermissions(zip_path, "r") as zip_ref:
        zip_ref.extractall(path)
    zip_path.unlink()
    version_tag.write_text(version_string)

    return exe_path


async def get_chrome(
    arch: str | None = None,
    i: int | None = None,
    path: str | Path = default_download_path,
    *,
    verbose: bool = False,
    force: bool = False,
) -> Path | str:
    """
    Download google chrome from google-chrome-for-testing server.

    Args:
        arch: the target platform/os, as understood by google's json directory.
        i: the chrome version: -1 being the latest version, 0 being the oldest
           still in the testing directory.
        path: where to download it too (the folder).
        verbose: print out version found
        force: download chrome again even if already present at that version

    """
    loop = asyncio.get_running_loop()
    fn = partial(
        get_chrome_sync,
        arch=arch,
        i=i,
        path=path,
        verbose=verbose,
        force=force,
    )
    return await loop.run_in_executor(
        executor=None,
        func=fn,
    )


def get_chrome_cli() -> None:
    if "ubuntu" in platform.version().lower():
        warnings.warn(  # noqa: B028
            "You are using `get_browser()` on Ubuntu."
            " Ubuntu is **very strict** about where binaries come from."
            " While sandbox is already off by default, do not set"
            " enable_sandbox to True OR you must install from Ubuntu's"
            " package manager.",
            UserWarning,
        )
    parser = argparse.ArgumentParser(
        description="Will download Chrome for testing. All arguments optional.",
        parents=[logistro.parser],
    )
    parser.add_argument(
        "--i",
        "-i",
        type=int,
        dest="i",
        help=(
            "Google offers thousands of chrome versions for download. "
            "'-i 0' is the oldest, '-i -1' is the newest: array syntax"
        ),
    )
    parser.add_argument(
        "--arch",
        dest="arch",
        help="linux64|win32|win64|mac-x64|mac-arm64",
    )
    parser.add_argument(
        "--path",
        dest="path",
        help="Where to store the download.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Display found version number if using -i (to stdout)",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Force download even if already present.",
    )
    parser.set_defaults(path=default_download_path)
    parser.set_defaults(arch=None)
    parser.set_defaults(verbose=False)
    parsed = parser.parse_args()
    i = parsed.i
    arch = parsed.arch
    path = Path(parsed.path)
    force = parsed.force
    verbose = parsed.verbose
    print(get_chrome_sync(arch=arch, i=i, path=path, verbose=verbose, force=force))  # noqa: T201 allow print in cli



from __future__ import annotations

import queue
import threading
from concurrent.futures import Executor, Future
from typing import TYPE_CHECKING

import logistro

if TYPE_CHECKING:
    from typing import Any, Callable, TypeVar

    try:
        from typing import ParamSpec
    except ImportError:
        from typing_extensions import ParamSpec

    _P = ParamSpec("_P")  # Runtime special generic that gives you access to fn sig
    _T = TypeVar("_T")


_logger = logistro.getLogger(__name__)


class ExecutorClosedError(RuntimeError):
    """Raise if submitting when executor is closed."""


class ManualThreadExecutor(Executor):
    def __init__(
        self,
        *,
        max_workers: int = 2,
        daemon: bool = True,
        name: str = "manual-exec",
    ) -> None:
        self._q: queue.Queue[
            tuple[  # could be typed more specifically if singleton @ submit()
                Callable[..., Any],
                Any,
                Any,
                Future[Any],
            ]
            | None
        ] = queue.Queue()
        self._stop = False
        self._threads = []
        self.name = name
        for i in range(max_workers):
            t = threading.Thread(
                target=self._worker,
                name=f"{name}-{i}",
                daemon=daemon,
            )
            t.start()
            self._threads.append(t)

    def _worker(self) -> None:
        while True:
            item = self._q.get()
            if item is None:  # sentinel
                return
            fn, args, kwargs, fut = item
            if fut.set_running_or_notify_cancel():
                try:
                    res = fn(*args, **kwargs)
                except BaseException as e:  # noqa: BLE001 yes we catch and set
                    fut.set_exception(e)
                else:
                    fut.set_result(res)
            self._q.task_done()

    # _T is generic so we can mar
    def submit(
        self,
        fn: Callable[_P, _T],
        /,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> Future[_T]:
        fut: Future[_T] = Future()
        if self._stop:
            fut.set_exception(ExecutorClosedError("Cannot submit tasks."))
            return fut
        self._q.put((fn, args, kwargs, fut))
        return fut

    def shutdown(
        self,
        wait: bool = True,  # noqa: FBT001, FBT002 overriding, can't change args
        *,
        cancel_futures: bool = False,
    ) -> None:
        self._stop = True
        if cancel_futures:
            # Drain queue and cancel pending
            try:
                while True:
                    full = self._q.get_nowait()
                    if full is not None:
                        _, _, _, fut = full
                        fut.cancel()
                    self._q.task_done()
            except queue.Empty:
                pass
        for _ in self._threads:
            self._q.put(None)
        if wait:
            for t in self._threads:
                t.join(timeout=2.0)



"""Provides a channel based on operating system file pipes."""

from __future__ import annotations

import os
import platform
import sys
import time
import warnings
from threading import Lock
from typing import TYPE_CHECKING

import logistro

from . import _wire as wire
from ._errors import BlockWarning, ChannelClosedError, JSONError

if TYPE_CHECKING:
    from typing import Any, Mapping, Sequence

    from choreographer.protocol import BrowserResponse

_with_block = bool(sys.version_info[:3] >= (3, 12) or platform.system() != "Windows")

_logger = logistro.getLogger(__name__)

# should be closing my ends from the start?


# if we're a pipe we expect these public attributes
class Pipe:
    """Defines an operating system pipe."""

    from_external_to_choreo: int
    """Consumers need this, it is the channel the browser uses to talk to choreo."""
    from_choreo_to_external: int
    """Consumers needs this, it is the channel choreo writes to the browser on."""
    shutdown_lock: Lock
    """Once this is locked, the pipe is closed and can't be reopened."""

    def __init__(self) -> None:
        """Construct a pipe using os functions."""
        # This is where pipe listens (from browser)
        # So pass the write to browser
        self._read_from_browser, self._write_from_browser = list(os.pipe())

        # This is where pipe writes (to browser)
        # So pass the read to browser
        self._read_to_browser, self._write_to_browser = list(os.pipe())

        # Popen will write stdout of wrapper to this (dupping 4)
        # Browser will write directly to this if not using wrapper
        self.from_external_to_choreo = self._write_from_browser
        # Popen will read this into stdin of wrapper (dupping 3)
        # Browser will read directly from this if not using wrapper
        # which dupes stdin to expected fd (4?)
        self.from_choreo_to_external = self._read_to_browser
        # These won't be used on windows directly, they'll be t-formed to
        # windows-style handles. But let another layer handle that.

        # this is just a convenience to prevent multiple shutdowns
        self.shutdown_lock = Lock()  # should be private
        self._open_lock = Lock()  # should be private

    def is_ready(self) -> bool:
        """Return true if pipe open."""
        return not self.shutdown_lock.locked() and self._open_lock.locked()

    def open(self) -> None:
        """
        Open the channel.

        In a sense, __init__ creates the pipe. The OS opens it.
        Here we're just marking it open for use, that said.

        We only use locks here for indications, we never actually lock,
        because the broker is in charge of all async/parallel stuff.
        """
        if not self._open_lock.acquire(blocking=False):
            raise RuntimeError("Cannot open same pipe twice.")

    def write_json(self, obj: Mapping[str, Any]) -> tuple[float, float]:
        """
        Send one json down the pipe.

        Args:
            obj: any python object that serializes to json.

        """
        if not self.is_ready():
            raise ChannelClosedError(
                "The communication channel was either never "
                "opened or closed. Was .open() or .close() called?",
            )
        encoded_message = wire.serialize(obj) + b"\0"
        _logger.debug(
            f"Writing message {encoded_message[:15]!r}...{encoded_message[-15:]!r}, "
            f"size: {len(encoded_message)}.",
        )
        _logger.debug2(f"Full Message: {encoded_message!r}")
        start = time.perf_counter()
        try:
            ret = os.write(self._write_to_browser, encoded_message)
            _logger.debug(
                f"***Wrote {ret}/{len(encoded_message)}***",
            )
            if ret != len(encoded_message):
                _logger.critical(
                    f"***Did not write entire message. {ret}/{len(encoded_message)}***",
                )
        except OSError as e:
            self.close()
            raise ChannelClosedError from e
        return (start, time.perf_counter())

    def read_jsons(  # noqa: PLR0912, PLR0915, C901 branches, complexity
        self,
        *,
        blocking: bool = True,
    ) -> tuple[Sequence[BrowserResponse], float]:
        """
        Read from the pipe and return one or more jsons in a list.

        Args:
            blocking: The read option can be set to block or not.

        Returns:
            A list of jsons.

        """
        jsons: list[BrowserResponse] = []
        if not self.is_ready():
            raise ChannelClosedError(
                "The communication channel was either never "
                "opened or closed. Was .open() or .close() called?",
            )
        if not _with_block and not blocking:
            warnings.warn(  # noqa: B028
                "Windows python version < 3.12 does not support non-blocking",
                BlockWarning,
            )
        try:
            if _with_block:
                os.set_blocking(self._read_from_browser, blocking)
        except OSError as e:
            self.close()
            raise ChannelClosedError from e
        raw_buffer = None  # if we fail in read, we already defined
        loop_count = 1
        try:
            raw_buffer = os.read(
                self._read_from_browser,
                10000,
            )  # 10MB buffer, nbd, doesn't matter w/ this
            _logger.debug(
                f"First read in loop: {raw_buffer[:15]!r}...{raw_buffer[-15:]!r}. "
                f"size: {len(raw_buffer)}.",
            )
            _logger.debug2(f"Whole buffer: {raw_buffer!r}")
            if not raw_buffer or raw_buffer == b"{bye}\n":
                if raw_buffer:
                    _logger.debug(f"Received {raw_buffer!r}. is bye?")
                # we seem to need {bye} even if chrome closes NOTE
                self.close()
                raise ChannelClosedError
            while raw_buffer[-1] != 0:
                _logger.debug("Partial message from browser received.")
                loop_count += 1
                if _with_block:
                    os.set_blocking(self._read_from_browser, True)
                raw_buffer += os.read(self._read_from_browser, 10000)
        except BlockingIOError:
            _logger.debug("BlockingIOError")
            return jsons, time.perf_counter()
        except OSError as e:
            _logger.debug("OSError")
            self.close()
            if not raw_buffer or raw_buffer == b"{bye}\n":
                raise ChannelClosedError from e
            # this could be hard to test as it is a real OS corner case
        finally:
            _logger.debug(
                f"Total loops: {loop_count}, "
                f"Final size: {len(raw_buffer) if raw_buffer else 0}.",
            )
            _logger.debug2(f"Whole buffer: {raw_buffer!r}")
        if raw_buffer is None:
            return jsons, time.perf_counter()
        decoded_buffer = raw_buffer.decode("utf-8")
        raw_messages = decoded_buffer.split("\0")
        _logger.debug(f"Received {len(raw_messages)} raw_messages.")
        for raw_message in raw_messages:
            if raw_message:
                try:
                    jsons.append(wire.deserialize(raw_message))
                except JSONError:
                    _logger.exception("JSONError decoding message. Ignoring")
                except:
                    _logger.exception("Error in trying to decode JSON off our read.")
                    raise
        return jsons, time.perf_counter()

    def _unblock_fd(self, fd: int) -> None:
        try:
            if _with_block:
                os.set_blocking(fd, False)
        except Exception:  # noqa: BLE001, S110 OS errors are not consistent, catch blind + pass
            pass

    def _close_fd(self, fd: int) -> None:
        try:
            os.close(fd)
        except Exception:  # noqa: BLE001, S110 OS errors are not consistent, catch blind + pass
            pass

    def _fake_bye(self) -> None:
        self._unblock_fd(self._write_from_browser)
        try:
            os.write(self._write_from_browser, b"{bye}\n")
        except Exception:  # noqa: BLE001, S110 OS errors are not consistent, catch blind + pass
            pass

    def close(self) -> None:
        """Close the pipe."""
        if self.shutdown_lock.acquire(blocking=False):
            if platform.system() == "Windows":
                self._fake_bye()
            self._unblock_fd(self._write_from_browser)
            self._unblock_fd(self._read_from_browser)
            self._unblock_fd(self._write_to_browser)
            self._unblock_fd(self._read_to_browser)
            self._close_fd(self._write_to_browser)  # no more writes
            self._close_fd(self._write_from_browser)  # we're done with writes
            self._close_fd(self._read_from_browser)  # no more attempts at read
            self._close_fd(self._read_to_browser)



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import logistro
import simplejson
from simplejson import errors as sjerrors

from ._errors import JSONError

if TYPE_CHECKING:
    from typing import Any

_logger = logistro.getLogger(__name__)

_custom_encoder: type[json.JSONEncoder] | None = None


def register_custom_encoder(e: type[json.JSONEncoder] | None) -> None:
    global _custom_encoder  # noqa: PLW0603 what other choice do we have
    _custom_encoder = e


class MultiEncoder(simplejson.JSONEncoder):
    """Special json encoder for numpy types."""

    # docs say subclassing inferior, just pass this method as a kward
    def default(self, o: Any) -> Any:
        if hasattr(o, "dtype") and o.dtype.kind in ("i", "u") and o.shape == ():
            return int(o)
        elif hasattr(o, "dtype") and o.dtype.kind == "f" and o.shape == ():
            return float(o)
        elif hasattr(o, "dtype") and o.shape != ():
            if hasattr(o, "values") and hasattr(o.values, "tolist"):
                return o.values.tolist()
            if hasattr(o, "tolist"):
                return o.tolist()
        elif hasattr(o, "isoformat"):
            return o.isoformat()
        return simplejson.JSONEncoder.default(self, o)


def serialize(obj: Any) -> bytes:
    try:
        if not _custom_encoder:
            message = simplejson.dumps(
                obj,
                ensure_ascii=False,
                ignore_nan=True,
                cls=MultiEncoder,
            )
        else:
            message = json.dumps(obj, cls=_custom_encoder)
    except (json.JSONDecodeError, simplejson.JSONDecodeError) as e:
        raise JSONError from e
    _logger.debug(f"Serialized: {message[:15]}...{message[-15:]}, size: {len(message)}")
    _logger.debug2(f"Whole message: {message}")

    return message.encode("utf-8")


def deserialize(message: str) -> Any:
    try:
        return simplejson.loads(message)
    except sjerrors.JSONDecodeError as e:
        raise JSONError from e
